/*
 * lcm_server.c - Daemon for Allnet FW8888 LCM control
 *
 * Copyright 2010 ALLNET GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/time.h>

#define MAX_FILENAME_LENGTH 16
#define RESULT_OK 0x11
#define RESULT_FAIL 0x14
#define RESULT_LENGTH 16
#define CMD_MAX_LENGTH 32
#define STR_MAX_LENGTH 16
#define SELECT_TIMEOUT 10 //in seconds

/**
 * useconds before get command results from LCM
 */
#define timing_val 100000

/**
 * Command FIFO to daemon
 */
char cmd_pipe[]="/var/run/lcm_cmd";

/**
 * Key FIFO from daemon
 */
char key_pipe[]="/var/run/lcm_key";

/**
 * Commands to LCM
 */
char Reset_LCM[]={0x10,0x39};
char Reset_LCM1[]={0x10,0x0,0x39}; //special command for alignment fix
char Clear_display[]={0x10,0x01};
char Return_home[]={0x10,0x02};
char Display_ON[]={0x10,0x0E};
char Display_ON_NOCURSOR[]={0x10,0x0C};
char Display_OFF[]={0x10,0x08};
char Bklight_OFF[]={0x10,0x38};
char Bklight_ON[]={0x10,0x39};

/**
 * Command sets to Daemon
 */
char CMD_clear[]="Clear";
char CMD_home[]="Home";
char CMD_disp[]="Display";
char CMD_bklight[]="BKLight";
char CMD_setpos[]="Setpos";
char CMD_write[]="Write";
char CMD_on[]="on";
char CMD_off[]="off";
char CMD_nocur[]="nocur";

/**
 *	print_buf - utility function for buffer hex dump
 *	@buf: buffer for dump 
 *	@count: content length in buffer
 *
 *      Return: none
 */
void print_buf(char *buf, int count)
{
  int i;
  if(count>RESULT_LENGTH)
    count=RESULT_LENGTH;

  printf("Buf[%d]=",count);
  for(i=0;i<count;i++)
    printf("0x%02X,",buf[i]);

  printf("\n");
}

/**
 *	read_lcm - read command result and key from LCM
 *	@fd: serial file descriptor 
 *	@read_max: content length to read
 *
 *      Return: 
 *              0 for nothing
 *              other values as input
 *
 *      Note: Only return the first byte, others discard
 *            
 */
char read_lcm(int fd, int read_max)
{
  char ret;
  int read_count;
  char buf[RESULT_LENGTH];

  usleep(timing_val); //adjust read timing
  read_count=read(fd, buf, read_max);
  if(read_count>0){
    ret=buf[0];
    return ret;
  }
  else
    return 0;
}

/**
 *	reset_lcm - reset the LCM to initial state if communication error
 *	@fd: serial file descriptor 
 *
 *      Return: none
 */
void reset_lcm(int fd)
{
  char buf[RESULT_LENGTH];
  int read_count;

  write(fd,Reset_LCM,sizeof(Reset_LCM));
  sleep(1);
  read_count=read(fd, buf,RESULT_LENGTH);
  printf("Reset command return %d bytes\n",read_count);

  /**
   *    Reset alignment is required if more than 1 ack received
   */
  if(read_count>=1){
    printf("Reset command return[%d]: %02X\n",read_count,buf[0]);
    if(buf[0]==0x14){
	write(fd,Reset_LCM1,sizeof(Reset_LCM1));
        read_count=read(fd, buf,RESULT_LENGTH);
  	if(read_count>=1)
    	  printf("Reset align command return[%d]: %02X %02X\n",read_count,buf[0],buf[1]);
    }	
  }
}

/**
 *	init_lcm - initialize the communication to LCM 
 *	@fd: serial file descriptor 
 *
 *      Return: serial file descriptor
 */
int init_lcm(char * device_file)
{
  int fd;
  struct termios options;
  int read_count;
  char buf[RESULT_LENGTH];
  memset((void *)&options, 0,sizeof(options));

  printf("Using serial device: %s\n",device_file);

  fd = open(device_file, O_RDWR | O_NOCTTY |O_SYNC);
  if(fd < 0){
    printf("Error opening device %s, exiting\n",device_file);
    return -1;
  }
  tcdrain(fd);

  /**
   *    Set the communication parameters
   */
  options.c_cflag &= ~(PARENB | CSTOPB );
  options.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
  options.c_iflag |= (IGNPAR | ECHOCTL | ECHOKE );
  options.c_lflag &= ~(ICANON | OPOST| ECHO | ECHOE | ISIG | IEXTEN | ECHOK | ONLCR |IXON );
  options.c_cc[VMIN]  = 0;
  options.c_cc[VTIME]  = 0;

  tcsetattr(fd, TCSAFLUSH, &options);

  /**
   *    Issue a reset to LCM 
   */
  reset_lcm(fd);

  return fd;
}

/**
 *	process_cmd - process the commands from command FIFO
 *	@com_fd: serial file descriptor 
 *	@cmd_fd: command FIFO file descriptor
 *
 *      Return: none
 *
 *      Note: none
 *            
 */

void process_cmd(int com_fd, int cmd_fd){
  char buf[CMD_MAX_LENGTH];
  char str_buf[2];
  
  int x,y,nread,i;
  char *next_buf,ret;

  memset(buf,0,CMD_MAX_LENGTH);

  /**
   *    Read commands from command FIFO 
   */
  nread=read(cmd_fd,buf,CMD_MAX_LENGTH);
  //printf("Got CMD[%d]:%s\n",nread,buf);

  /**
   *    Check for "Clear" command
   */
  if(strncmp(buf,CMD_clear,strlen(CMD_clear))==0){
    write(com_fd,Clear_display,sizeof(Clear_display));
  }

  /**
   *    Check for "Home" command
   */
  if(strncmp(buf,CMD_home,strlen(CMD_home))==0){
    write(com_fd,Return_home,sizeof(Return_home));
  }

  /**
   *    Check for "Display" command
   *    Parameter: "on", "off", "nocur"
   */
  if(strncmp(buf,CMD_disp,strlen(CMD_disp))==0){
    for(i=strlen(CMD_disp);i<nread;i++)
      if(buf[i]!=' ')
	break;
    if(i==nread){
      printf("Command no parameter\n");
      return;
    }
    next_buf=&buf[i];
    if(strncmp(next_buf,CMD_on,strlen(CMD_on))==0)
      write(com_fd,Display_ON,sizeof(Display_ON));
    else if(strncmp(next_buf,CMD_off,strlen(CMD_off))==0)
      write(com_fd,Display_OFF,sizeof(Display_OFF));
    else if(strncmp(next_buf,CMD_nocur,strlen(CMD_nocur))==0)
      write(com_fd,Display_ON_NOCURSOR,sizeof(Display_ON_NOCURSOR));
    else
      ;
 
  }


  /**
   *    Check for "BKLight" command, control the backlight
   *    Parameter: "on", "off"
   */
  if(strncmp(buf,CMD_bklight,strlen(CMD_bklight))==0){
    for(i=strlen(CMD_bklight);i<nread;i++)
      if(buf[i]!=' ')
	break;
    if(i==nread){
      printf("Command no parameter\n");
      return;
    }
    next_buf=&buf[i];
    if(strncmp(next_buf,CMD_on,strlen(CMD_on))==0){
      write(com_fd,Bklight_ON,sizeof(Bklight_ON));
    }
    else if(strncmp(next_buf,CMD_off,strlen(CMD_off))==0){
     write(com_fd,Bklight_OFF,sizeof(Bklight_OFF));
    }
    else
      ;
  }

  /**
   *    Check for "Setpos" command, control the cursor position for next appearence
   *    Parameter #1: row address -- 0, 1
   *    Parameter #2: column address -- 0~15
   */
  if(strncmp(buf,CMD_setpos,strlen(CMD_setpos))==0){
     for(i=strlen(CMD_setpos);i<nread;i++)
      if(buf[i]!=' ')
	break;
    if(i==nread){
      printf("Command no parameter\n");
      return;
    }
    next_buf=&buf[i];
    sscanf(next_buf,"%d %d",&x,&y);
    //printf("Next buf=%s, X=%d, Y=%d\n",next_buf,x,y);
    buf[0]=0x10;
    if(x==0)
      buf[1]=0x80+y; //0x80 is the base address of row 0
    else
      buf[1]=0xc0+y; //0xc0 is the base address of row 1

    write(com_fd,buf,2);
  }

  /**
   *    Check for "write" command to output string
   *    Parameter: string to write (will be cut if exceed display boundary)
   *
   *    Note: Tailing null character of string will not be output
   */
  if(strncmp(buf,CMD_write,strlen(CMD_write))==0){
     for(i=strlen(CMD_setpos);i<nread;i++)
      if(buf[i]!=' ')
	break;
    if(i==nread){
      printf("Command no parameter\n");
      return;
    }
    str_buf[0]=0x12;  //fixed write command
    for(;i<nread-1;i++){
      str_buf[1]=buf[i];
      //printf("Buf[%d]=0x%02X\n",i,buf[i]);
      write(com_fd,str_buf,2);
    }
  }

  //  usleep(timing_val);
  ret=read_lcm(com_fd,RESULT_LENGTH);
  if(ret!=RESULT_OK){
    printf("Ret=0x%02X\n",ret);
    reset_lcm(com_fd);
  }
  return;
  

}

/**
 *	main - main loop of daemon
 *	@argv[1]: serial device file "/dev/ttyS?" 
 *	@argv[2]: quoted string to display while daemon start
 *
 *      Return:   -1 for error
 *              
 *
 *            
 */

int main(int argc, char **argv)
{
  int fd,cmd_fd,key_fd;
  int i, toggle=0;
  char buf[RESULT_LENGTH];
  char title[STR_MAX_LENGTH*2];
  int title_length;
  int read_count;

  int retval;
  char com_dev[MAX_FILENAME_LENGTH];
  char ret;
  
  fd_set read_fds;
  struct timeval tv;

  tv.tv_sec=SELECT_TIMEOUT;
  tv.tv_usec=0;

  memset((void *)com_dev,0,MAX_FILENAME_LENGTH);
  memset((void *)title,0,STR_MAX_LENGTH*2);
  
  if(argc <3){
    printf("Usage: %s [/dev/ttySx] \"Title\"\n",argv[0]);
    return -1;
  }
  else{
    strncpy((char *)com_dev,argv[1],MAX_FILENAME_LENGTH-1);

  /**
   *    Fill the output buffer for title display.
   *    Even bytes for LCM write command 0x12, odd bytes for characters
   */
    for(i=0;i<STR_MAX_LENGTH;i++){
      if(!argv[2][i])
	break;
      title[i*2]=0x12;
      title[i*2+1]=argv[2][i];
    }
    title_length=i;
  }
  

  FD_ZERO(&read_fds);
 
  fd=init_lcm(com_dev);
  if(fd <0)
    return -1;
  
  FD_SET(fd,&read_fds);

   /**
   *    Create command and key FIFOs
   *    Exit if failed
   */
  if(mkfifo(cmd_pipe,S_IRWXU)&&(errno!=EEXIST)){
    printf("Cannot create command pipe: %s\n",cmd_pipe);
    return -1;
  }
  if(mkfifo(key_pipe,S_IRWXU)&&(errno!=EEXIST)){
    printf("Cannot create key pipe: %s\n",key_pipe);
    return -1;
  }

   /**
   *    Open with O_RDWR to avoid EOF condition on select
   *   
   */
  cmd_fd=open(cmd_pipe,O_RDWR|O_NDELAY); 
  if(cmd_fd<0){
    printf("Cannot open command pipe: %s\n",cmd_pipe);
    return -1;
  }

   /**
   *    Flush previous commands in FIFO
   */
  ret=read(cmd_fd,buf,16);
  if(ret>1)
    printf("Flushing commands[%d]: %s\n",ret,buf);
  
  FD_SET(cmd_fd,&read_fds);

  key_fd=open(key_pipe,O_RDWR|O_NDELAY);
  if(key_fd<0){
    printf("Cannot open key pipe: %s\n",key_pipe);
    return -1;
  }

   /**
   *    Flush previous keys in FIFO
   */
  ret=read(key_fd,buf,16);
  if(ret>1)
    printf("Flushing key [%d]: %s\n",ret,buf);

   /**
   *    Clear LCM display
   */
  write(fd,Clear_display,sizeof(Clear_display));
  ret=read_lcm(fd,RESULT_LENGTH);
  if(ret!=RESULT_OK)
    reset_lcm(fd);
  
   /**
   *    Set cursor back to 0,0
   */
  write(fd,Return_home,sizeof(Return_home));
  ret=read_lcm(fd,RESULT_LENGTH);
  if(ret!=RESULT_OK)
    reset_lcm(fd);

   /**
   *    Turn on the LCM display and hide cursor
   */
  write(fd,Display_ON_NOCURSOR,sizeof(Display_ON_NOCURSOR));
  ret=read_lcm(fd,RESULT_LENGTH);
  if(ret!=RESULT_OK)
    reset_lcm(fd);

  
  usleep(timing_val);

   /**
   *    Output title text at 0,0
   */
  write(fd,title,title_length*2);
  usleep(timing_val);
  read_count=read(fd, buf,RESULT_LENGTH);
  if(read_count>0)
    printf("Write command return[%d]: %02X\n",read_count,buf[0]);
  
  i=0;

   /**
   *    Main loop
   */
  while(1){

    retval=select(cmd_fd+1,&read_fds,NULL,NULL,&tv);
    if(retval>0){ //There are fds ready for read
      if(FD_ISSET(cmd_fd,&read_fds)){ // Command FIFO
 	process_cmd(fd,cmd_fd);

	FD_SET(cmd_fd,&read_fds);
      }  
      if(FD_ISSET(fd,&read_fds)){ //Serial port events
	ret=read_lcm(fd,1);
	memset(buf,0,RESULT_LENGTH);
	switch(ret){
	case 0x15:
	  sprintf(buf,"UP\n");
	  break;
	case 0x16:
	  sprintf(buf,"DOWN\n");
	  break;
	case 0x17:
	  sprintf(buf,"RET\n");
	  break;
	case 0x18:
	  sprintf(buf,"ESC\n");
	  break;
	default:
	  printf("Got meanless response from LCM: 0x%02X\n",ret);
	}

	/**
	 *    Output to key FIFO if key available
	 */
	if(buf[0]!=0)
	  write(key_fd,buf,strlen(buf));
	FD_SET(cmd_fd,&read_fds);
      }  
     }
     else if(retval<0){
       printf("Select error: %x\n",errno);
     }
     else  //select timeout
       ;

     tv.tv_sec=SELECT_TIMEOUT;
     tv.tv_usec=0;
     FD_ZERO(&read_fds);
     FD_SET(fd,&read_fds);
     FD_SET(cmd_fd,&read_fds);
    i++;
  }

  close(fd);
  close(cmd_fd);
  close(key_fd);
  return 0;
}
